/**
 CS 463 midterm exam code for problem 2:
   Floating point elliptic curve computations.

 Dr. Orion Lawlor, lawlor@alaska.edu, 2015-04-27 (Public Domain)
*/
#include <iostream>
#include <stdlib.h>
#include <cmath>
using std::fmod;

/* This float type represents our elliptic curve coordinates. */
typedef double ECcoord;

/* 2 is our prime */
ECcoord P=(ECcoord)2;

ECcoord modP(ECcoord in) {
	ECcoord out=fmod(in,P);
	while (out<0.0) out+=P; // wrap to positive image (fmod returns negative output for negative input)
	return out;
}

class ECcurve; // forward declaration

// One elliptic curve point, represented by 2D coordinates (x,y).
class ECpoint {
public:
   ECcoord x,y; // coordinates of point on the curve
   
   // Normal initialization
   ECpoint(const ECcoord &x_,const ECcoord &y_) :x(x_), y(y_) {}
   
   // By default, points are initialized to infinity.
   ECpoint() :x(infinity.x), y(infinity.y) {}
   
   // This is the "point at infinity".  It's arbitrarily far up.
   //   In the group of ECpoints, it's the identity element.
   static ECpoint infinity; 
   
   friend std::ostream &operator<<(std::ostream &o,const ECpoint &p) {
   	if (p == ECpoint::infinity) o<<"[infinity]";
   	else o<<p.x<<" , "<<p.y;
	return o;
   }
   
   bool operator==(const ECpoint &p) const {
      return fabs(x-p.x)<1.0e-6 && fabs(y-p.y)<1.0e-6;
   }
   bool operator!=(const ECpoint &p) const {
      return !operator==(p);
   }
   
 
   // Add us to o, under the action of this curve.
   ECpoint add(const ECpoint &o,const ECcurve &curve) const;
   
   // Multiply us by this scalar, under the action of this curve.
   ECpoint multiply(long scalar,const ECcurve &curve) const;
};

/******************* curves *************************/
class ECcurve {
public:
    // We represent an elliptic curve of the form
    //   y^2 = x^3 + a*x + b
    ECcoord a,b;

   // Here's a (arbitrary) starting point on the curve.
   //  This is called a curve generator G.
    ECpoint start;
    
    ECcurve() { 
    	a=(ECcoord)0.12345678; 
    	b=(ECcoord)0.7; 
    	start.x=(ECcoord)0.3141593;
    	start.y=(ECcoord)sqrt(evaluate(start.x,0.0)); 
    	std::cout<<"Generator point ="<<start<<"\n";
    }

    // Evaluate the elliptic curve at this point. 
    //  Returns 0 if the point lies along our curve.
    ECcoord evaluate(const ECcoord &x,const ECcoord &y) const {
    	return modP(y*y-(x*x*x+a*x+b));
    }
    bool on_curve(const ECpoint &p) const {
    	ECcoord e=evaluate(p.x,p.y);
    	if (e<1.0e-8 || e>P-1.0e-8) return true;
    	else return false;
    }

    // Evaluate the slope of the tangent of our curve at this curve point.
    ECcoord tangent(const ECcoord &x,const ECcoord &y) const {
    	ECcoord dx=modP(3*x*x+a);
    	ECcoord dy=2*y;
    	return modP(dx/dy);
    }
    
};


/* The elliptic curve point "infinity" is the additive identity (0) */
ECpoint ECpoint::infinity(-1,-1);

ECpoint ECpoint::add(const ECpoint &o,const ECcurve &curve) const
{
	// special cases for infinity:
	if (*this == infinity) return o; /* we're zero */
	if (o == infinity) return *this; /* they're zero */
	
	// std::cout<<"Adding curve points "<<o<<" and "<<*this<<"\n";

	// Usual case:
	ECcoord m;
	if (x-o.x==0) { // special case for same X coordinate
		if (y-o.y==0) { // adding point to itself--take tangent
			m=curve.tangent(x,y);
		} else { // adding additive inverses--return "infinity"
			return infinity;
		}
	}
	else { // default case: 
		m=modP((y-o.y) / (x-o.x)); // dy/dx slope of line (default case)
	}

// std::cout<<"EC slope m="<<m<<"\n";

	ECcoord v = modP(y-m*x); // line's y intercept

	ECcoord x3 = modP(m*m - x - o.x); // comes from matching x^2 terms
	ECcoord y3 = modP(-(m*x3+v)); // from line equation, plus mirroring

	return ECpoint(x3,y3);
}

// Multiply us by this scalar, under the action of this curve.
ECpoint ECpoint::multiply(long scalar,const ECcurve &curve) const
{
	ECpoint sum=ECpoint::infinity; 
	for (int copy=0;copy<scalar;copy++)
		sum=sum.add(*this,curve);
	return sum;
}


int main(int argc,char *argv[]) {
/* Agree on a curve */
  ECcurve curve;
  for (long test=0;test<2000;test++) {
  	long magnitude=1000000+test;

/* Prepare local secrets (simple HACK: hardcoded, not random) */
	ECcoord a=1234567890%magnitude;
	ECcoord b=1987654321%magnitude;
	
/* Move along the curve to create public points */
	ECpoint A=curve.start.multiply(a,curve);
	ECpoint B=curve.start.multiply(b,curve);

/* Create both combinations of public keys */
	ECpoint AB=A.add(B,curve); 
	ECpoint BA=B.add(A,curve); 
	
	std::cout<<test<<"			AB="<<AB<<" BA="<<BA<<"\n";
	
/* Compare points */
	if (fabs(AB.x-BA.x)<1.0e-6) {
		std::cout<<"OK, shared secret matches.\n";
	} else {
		std::cout<<"Problem: shared secrets are different!\n";
		exit(1);
	}	
  }
}


